"""A python package to dump any HLS stream as it is"""
import os.path
import sys

# It imports release module this way because if it tried to import the whole package
# and some required dependencies were not installed, that would fail
# This is the only way to access the release module without needing all
# dependencies.
sys.path.insert(0, os.path.dirname(__file__))
import release

__author__ = release.author
__version__ = release.version

# List of all modules that contain classes needed to be registered with the registry (via a Class.register decorator)
# This code ensures that the modules are imported and hence the decorators are
# executed and the classes/functions registered.
__modules_to_register = []

for m in [None] + __modules_to_register:
    # This try/catch is needed to support reload(dump_hls)
    if m is not None:
        try:
            exec('import %s' % m)
            exec('del %s' % m)
        except AttributeError:
            pass
else:
    if m is not None:
        del m



import asyncio
import logging
import os.path
import pathlib
from dataclasses import dataclass
from enum import Enum
from typing import Dict, List
from urllib.parse import urlparse, urljoin

import aiohttp
import joblib
import requests
from tqdm import tqdm

__all__ = [
    'HlsLine',
    'HlsLineType',
    'StreamDumper'
]

TAG_PREFIX = 'EXT'
DOWNLOAD_BLOCK_SIZE = 1024

logger = logging.getLogger('DumpHls')


class HlsTag(Enum):
    EXTINF = 'EXTINF'


class HlsLineType(Enum):
    TAG = 0
    URI = 1
    COMMENT = 2
    BLANK = 3


@dataclass
class HlsLine:
    type: HlsLineType = HlsLineType.BLANK
    tag: str = ''
    attributes: dict = None
    value: str = ''

    def __post_init__(self):
        if self.attributes is None:
            self.attributes = {}

    @staticmethod
    def _parse_attributes(attr_str):
        def clear_quote(value):
            if value.startswith('"'):
                return value[1:-1]
            return value

        attrs = {}
        l, attr_name, in_quote = 0, None, False
        for r, c in enumerate(attr_str):
            if c == '=':
                attr_name = attr_str[l:r]
                l = r + 1
            elif c == ',' and not in_quote:
                attrs[attr_name] = clear_quote(attr_str[l:r])
                attr_name = None
                l = r + 1
            elif c == '"':
                in_quote = not in_quote
        else:
            val = clear_quote(attr_str[l:])
            if len(val) > 0:
                attrs[attr_name] = val
        return attrs

    @staticmethod
    def parse(line):
        line = line.strip()
        res = HlsLine()
        if line.startswith('#'):
            line = line[1:]
            parts = line.split(':', maxsplit=1)
            if parts[0].startswith(TAG_PREFIX):
                res.type = HlsLineType.TAG
                res.tag = parts[0]
                if len(parts) > 1:
                    res.attributes = HlsLine._parse_attributes(parts[1])
            else:
                res.type = HlsLineType.COMMENT
                res.value = line
        elif len(line) > 0:
            res.type = HlsLineType.URI
            res.value = line
        else:
            res.type = HlsLineType.BLANK
        return res


class DownloadMode(Enum):
    SEQUENTIAL = 0
    ASYNC = 1
    MULTIPROCESS = 2


class StreamDumper:
    def __init__(self, parent_dir):
        if not isinstance(parent_dir, pathlib.Path):
            parent_dir = pathlib.Path(parent_dir)
        self.parent_dir = parent_dir

        # state
        self.known_domain_names = set()

    def reset(self):
        self.known_domain_names = set()

    def get_fs_path(self, url: str) -> pathlib.Path:
        """

        Args:
            url: this should be an absolute path including the domain name

        Returns:

        """
        parsed = urlparse(url)
        return self.parent_dir.joinpath(parsed.netloc, parsed.path.strip('/'))

    def download(self, url, dest: pathlib.Path, *, show_progress=False):
        res = requests.get(url, stream=show_progress)
        res.raise_for_status()
        total_size = int(res.headers.get('content-length', 0))

        dest.parent.mkdir(mode=0o755, parents=True, exist_ok=True)
        if show_progress:
            # TODO: fix multiple progress bar
            with open(dest, 'wb') as file, tqdm(
                    desc=str(dest.relative_to(self.parent_dir)),
                    total=total_size,
                    unit='iB',
                    unit_scale=True,
                    unit_divisor=1024,
            ) as bar:
                for data in res.iter_content(DOWNLOAD_BLOCK_SIZE):
                    bar.update(len(data))
                    file.write(data)
        else:
            dest.write_text(res.text)

    def download_file(self, url, *, update: bool = False, show_progress: bool = False) -> pathlib.Path:
        """Make sure the file is stored locally.

        Args:
            url: should be an absolute URL with the domain name
            update: if true, even when the file path has the content, the stream dumper will still download the file \
                    and update the local file.

        Returns:
            the path to the local file
        """
        fs_path = self.get_fs_path(url)
        if not fs_path.exists() or update:
            self.known_domain_names |= {urlparse(url).netloc}
            self.download(url, fs_path, show_progress=show_progress)
        assert fs_path.exists(), f"There should be a file under {fs_path.absolute()}"
        return fs_path

    @staticmethod
    def parse_playlist(payload: str, url: str) -> Dict[str, List[str]]:
        """Parse the playlist and get other files needed to be downloaded.

        Args:
            payload: the content of the playlist
            url: the url of this playlist

        Returns:
            Key ``playlists`` refers to all the playlist that needs parsing, and "files" refers to all the files \
            that need downloading.
        """
        lines = [HlsLine.parse(l) for l in payload.split('\n')]
        urls = []
        for line in lines:
            if line.type == HlsLineType.TAG:
                uri = line.attributes.get('URI', None)
                if uri is not None:
                    urls.append(uri)
            elif line.type == HlsLineType.URI:
                urls.append(line.value)

        playlists = set()
        files = set()
        for uri in urls:
            parsed = urlparse(uri)
            _, ext = os.path.splitext(parsed.path)
            if not (parsed.scheme in {'http', 'https'} or (parsed.netloc == '' and ext != '')): continue
            if parsed.netloc != '' and parsed.netloc != urlparse(url).netloc:
                logger.warning(f"Absolute URL under other domain name found: {uri}")
            full = urljoin(url, uri)
            if ext.lower() in {'.m3u8', '.m3u'}:
                playlists.add(full)
            else:
                files.add(full)
        logger.critical(f"{len(playlists)} playlists and {len(files)} files found.")
        return dict(playlists=list(sorted(playlists)), files=list(sorted(files)))

    async def download_and_parse_playlist_async(self, session: aiohttp.ClientSession, url: str, *,
                                                update: bool = False) -> Dict[str, List[str]]:
        """Make sure the file is stored locally.

        Args:
            session: the aiohttp session
            url: should be an absolute URL with the domain name
            update: if true, even when the file path has the content, the stream dumper will still download the file \
                    and update the local file.

        Returns:
            Key ``playlists`` refers to all the playlist that needs parsing, and "files" refers to all the files \
            that need downloading.
        """
        logger.critical(f"Downloading and Parsing: {url}")
        fs_path = self.get_fs_path(url)
        content = ''
        if not fs_path.exists() or update:
            self.known_domain_names |= {urlparse(url).netloc}
            async with session.get(url, ssl=False) as res:
                res.raise_for_status()
                content = await res.text()
                fs_path.parent.mkdir(mode=0o755, parents=True, exist_ok=True)
                fs_path.write_text(content)
        else:
            content = fs_path.read_text()
        return StreamDumper.parse_playlist(content, url)

    async def download_and_parse_playlists_async(self, urls, *, update=False):
        async with aiohttp.ClientSession() as session:
            tasks = []
            for url in urls:
                tasks.append(self.download_and_parse_playlist_async(session, url, update=update))
            return await asyncio.gather(*tasks)

    def download_and_parse_playlist(self, url, *, update=True):
        logger.critical(f'Downloading and Parsing: {url}')
        fs_path = self.download_file(url, update=update, show_progress=False)
        content = fs_path.read_text()
        return StreamDumper.parse_playlist(content, url)

    def download_and_parse_playlists(self, urls: List[str], *, update: bool = False, is_async: bool = False):
        if is_async:
            return asyncio.run(self.download_and_parse_playlists_async(urls, update=update))
        return [self.download_and_parse_playlist(url, update=update) for url in urls]

    def download_files(self, urls, *, update=False, n_proc=-1):
        joblib.Parallel(n_jobs=n_proc, verbose=1)(
            joblib.delayed(self.download_file)(url, update=update, show_progress=True) for url in urls)

    def dump(self, master_url, *, update=False, n_proc=-1):
        self.reset()

        files = set()

        # master
        media = self.download_and_parse_playlists([master_url], update=update, is_async=False)[0]
        files |= set(media['files'])

        # media
        segments = self.download_and_parse_playlists(media['playlists'], update=update)
        for segment in segments:
            files |= set(segment['files'])
            assert segment['playlists'] == [], "There should be no playlist found in the media playlist."
        self.download_files(files, update=update, n_proc=n_proc)



"""
This module contains all the information related to the current release of the
library including descriptions, version number, authors and contact
information.
"""
import sys

# module information
name = 'dump_hls'

# author information.
# Used by __init__, doc and setup
author = 'milliele'
author_email = 'milliele@qq.com'
maintainer = 'milliele'
maintainer_email = 'milliele@qq.com'

description_short = 'A python package to dump any HLS stream as it is'

# URL
url = 'https://github.com/milliele/DumpHls'

# ======================= gitRelease information =============================
version = '1.0.1'
"""Version information, used by __init__, doc, setup and gitRelease"""
release_name = f"v{version}"
"""The release title"""
release_body = """
1. v1.0.0 is incorrect on PyPI, upload a new version to fix the issue.
""".strip()
"""The release body"""
is_draft = False
"""Is this github release a draft"""
is_prerelease = False
"""Is this github release a prerelease"""


# ============================================================================

def validate_version(version):
    try:
        parts = version.split('-')
        main = parts[0]
        # validate main version
        try:
            major, minor, patch = main.split('.')
            assert ".".join(map(str, map(int, [major, minor, patch]))) == main
        except Exception as e:
            raise ValueError('Main version part should be like `major.minor.patch`. '
                             'Each part should be integer without leading 0.')
        # validate prerelease
        if len(parts) > 1:
            assert len(parts) == 2, ("There should at most be 1 `-` in the version name, "
                                     f"but {len(parts) - 1} are given.")
            try:
                pre_l, pre_n = parts[1].split('.')
            except Exception as e:
                raise ValueError('Prerelease part should be like `<pre_release_type>.<pre_release_number>`.')
            assert pre_l in {'alpha', 'beta'}, "Pre-release type could only be {'alpha', 'beta'}"
            assert pre_n.isdigit() and str(int(pre_n)) == pre_n, \
                "Pre-release number should be an integer without leading 0."
    except Exception as e:
        raise ValueError(f'Invalid version: `{version}`.\n'
                         f'{e}')


validate_version(version)

if __name__ == '__main__':
    if len(sys.argv) == 2:
        attr_name = sys.argv[1]
        if attr_name in globals():
            print(globals()[attr_name])



import os
import sys

from setuptools import setup

# It imports release module this way because if it tried to import the whole package
# and some required dependencies were not installed, that would fail
# This is the only way to access the release module without needing all
# dependencies.
sys.path.insert(0, 'src/dump_hls')
import release

sys.path.pop(0)

with open('README.md', 'r') as f:
    long_description = f.read()

with open('LICENSE', 'r') as f:
    license = f.read()

DEV_DEPENDENCIES = [
    'coverage',
    'twine',
    'pytest',
]


def parse_requirements_txt(file):
    with open('requirements.txt', "r") as f:
        lines = list(map(str.strip, f.readlines()))
    # remove comment
    lines = [l[:l.find('#')].strip() if '#' in l else l for l in lines if not l.startswith('#')]
    return lines


requirements_dev = parse_requirements_txt('requirements.txt')
requirements_prod = [r for r in requirements_dev if all(not r.startswith(p) for p in DEV_DEPENDENCIES)]

prefix = os.getenv("PYPI_PREFIX", '')
if prefix:
    prefix += '.'

setup(
    name=f'{prefix}{release.name}',
    version=release.version,
    author=release.author,
    author_email=release.author_email,
    description=release.description_short,
    long_description=long_description,
    long_description_content_type='text/markdown',
    url=release.url,
    include_package_data=True,
    package_dir={'': 'src'},
    packages=['dump_hls'],
    license=license,
    install_requires=requirements_prod,
    entry_points={
        'console_scripts': [
            'dumphls=dump_hls.cli:dumphls',
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
)



"""Run the Dump HLS.

This script defines the CLI of Dump HLS.

Check the instructions: ::

    dumphls -h
"""
import os.path
import sys
from urllib.parse import urlparse

import click

from main import StreamDumper
from . import __version__

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.version_option(__version__)
@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--server', type=click.Path(exists=True, file_okay=False), help='Directory path for server')
@click.argument('multivariant_url')
def dumphls(server: str, multivariant_url: str):
    """Dump HLS content from a given URL, optionally using a local server directory.

    Args:
        server: the directory path for the server where the files will be stored
        multivariant_url: URL to the multi-variant playlist (M3U8) file
    """
    click.echo(f"Server directory: {server}")
    click.echo(f"URL: {multivariant_url}")

    try:
        parsed_url = urlparse(multivariant_url)
    except Exception:
        click.echo(f"Invalid URL: {multivariant_url}", err=True)
        sys.exit(1)
    if not all([parsed_url.scheme in ('http', 'https'), parsed_url.netloc]):
        click.echo(f"URL must be an absolute URL with a domain name: {multivariant_url}", err=True)
        sys.exit(1)
    _, ext = os.path.splitext(parsed_url.path)
    if ext.lower() not in ('.m3u8', '.m3u'):
        click.echo(f"URL must point to a multi-variant playlist (M3U8) file: {multivariant_url}", err=True)
        sys.exit(1)
    d = StreamDumper(server)
    d.dump(multivariant_url)


if __name__ == '__main__':
    dumphls()
